# SPDX-License-Identifier: Apache-2.0
"""
This package contains the API for the OneLogger.

The classes in this package are used to set up OneLogger and capture spans, events, and errors.
See README.md for more information.
"""

from nv_one_logger.api.config import LoggerConfig, OneLoggerConfig, OneLoggerErrorHandlingStrategy  # noqa: F401
